import argparse
import logging
import os
import sys
import time
from dataclasses import dataclass
from typing import List, Tuple

import daiquiri
import requests
from dateutil.parser import parse
from slacker import Slacker


@dataclass(frozen=True)
class Message:
    user_id: str
    timestamp: float
    text: str

    @classmethod
    def from_dict(cls, data: dict) -> "Message":
        return cls(data["user"], float(data["ts"]), data["text"])


class SlackPurger:
    def __init__(self, token: str, max_timestamp: float) -> None:
        self.log = daiquiri.getLogger(self.__class__.__name__)
        self.max_timestamp = max_timestamp
        self._session = requests.Session()
        self._slack = Slacker(token, session=self._session)
        self.user_id, self.user_name = self.get_user()

    def get_user(self) -> Tuple[str, str]:
        response = self._slack.auth.test()
        return response.body["user_id"], response.body["user"]

    def parse_messages(self, messages: List[dict]) -> List[Message]:
        parsed_messages = [
            Message.from_dict(msg)
            for msg in messages
            if msg.get("user") is not None
        ]
        self.log.debug(f"Parsed {len(parsed_messages)} messages")
        return parsed_messages

    @property
    def channels(self) -> List[str]:
        self.log.debug("Listing channels")
        response = self._slack.channels.list()
        return [channel["id"] for channel in response.body["channels"]]

    def channel_history(self, channel_id: str) -> List[Message]:
        self.log.debug(f"Retrieving history for channel with ID {channel_id}")
        response = self._slack.channels.history(channel_id)
        return self.parse_messages(response.body["messages"])

    @property
    def groups(self) -> List[str]:
        self.log.debug("Listing groups")
        response = self._slack.groups.list()
        return [group["id"] for group in response.body["groups"]]

    def group_history(self, group_id: str) -> List[Message]:
        self.log.debug(f"Retrieving history for group with ID {group_id}")
        response = self._slack.groups.history(group_id)
        return self.parse_messages(response.body["messages"])

    @property
    def mpims(self) -> List[str]:
        self.log.debug("Listing MPIMs")
        response = self._slack.mpim.list()
        return [mpim["id"] for mpim in response.body["groups"]]

    def mpim_history(self, mpim_id: str) -> List[Message]:
        self.log.debug(f"Retreiving history for MPIM with ID {mpim_id}")
        response = self._slack.mpim.history(mpim_id)
        return self.parse_messages(response.body["messages"])

    @property
    def ims(self) -> List[str]:
        self.log.debug("Listing IMs")
        response = self._slack.im.list()
        return [im["id"] for im in response.body["ims"]]

    def im_history(self, im_id: str) -> List[Message]:
        self.log.debug(f"Retrieving history for IM with ID {im_id}")
        response = self._slack.im.history(im_id)
        return self.parse_messages(response.body["messages"])

    def delete_if_before_threshold(
        self, channel: str, message: Message
    ) -> None:
        if not self.should_delete(message):
            return

        try:
            time.sleep(1)
            self.log.info(f"Deleting message: {message.text}")
            self._slack.chat.delete(
                channel, str(message.timestamp), as_user=True
            )
        except Exception as exc:  # pylint: disable=broad-except
            self.log.warning(f"Error deleting message: {exc}")

    def should_delete(self, message: Message) -> bool:
        return (
            message.user_id is not None
            and message.user_id == self.user_id
            and message.timestamp < self.max_timestamp
        )

    def purge_channels(self) -> None:
        self.log.info(
            f"Purging messages in channels from user {self.user_name}"
        )
        for channel_id in self.channels:
            for message in self.channel_history(channel_id):
                self.delete_if_before_threshold(channel_id, message)

    def purge_groups(self) -> None:
        self.log.info(f"Purging messages in groups from user {self.user_name}")
        for group_id in self.groups:
            for message in self.group_history(group_id):
                self.delete_if_before_threshold(group_id, message)

    def purge_mpims(self) -> None:
        self.log.info(f"Purging messages in MPIMs from user {self.user_name}")
        for mpim_id in self.mpims:
            for message in self.mpim_history(mpim_id):
                self.delete_if_before_threshold(mpim_id, message)

    def purge_ims(self) -> None:
        self.log.info(f"Purging messages in IMs from user {self.user_name}")
        for im_id in self.ims:
            for message in self.im_history(im_id):
                self.delete_if_before_threshold(im_id, message)

    def purge_files(self) -> None:
        self.log.info(f"Purging files uploaded by user {self.user_name}")
        response = self._slack.files.list(
            ts_to=self.max_timestamp, user=self.user_id
        )
        for file_ in response.body["files"]:
            try:
                time.sleep(1)
                self.log.info(f"Deleting file {file_['id']}")
                self._slack.files.delete(file_["id"])
            except Exception as exc:  # pylint: disable=broad-except
                self.log.warning(f"Error deleting file: {exc}")

    def purge_all(self) -> None:
        self.purge_files()
        self.purge_channels()
        self.purge_groups()
        self.purge_mpims()
        self.purge_ims()


def cli() -> None:
    daiquiri.setup(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--end-time",
        required=True,
        type=lambda data: parse(data).timestamp(),
        help="Time before which messages will be purged",
    )

    if "SLACK_TOKEN" in os.environ:
        token_opts = dict(required=False, default=os.environ["SLACK_TOKEN"])
    else:
        token_opts = dict(required=True)

    parser.add_argument(  # type: ignore
        "--token", help="Slack API token", **token_opts
    )

    args = parser.parse_args()

    purger = SlackPurger(args.token, args.end_time)
    purger.purge_all()


def main() -> None:
    try:
        cli()
    except KeyboardInterrupt:
        sys.exit()


if __name__ == "__main__":
    main()
